import unittest

from pu import minimist


class Test(unittest.TestCase):
    def test_abc_1(self):
        self.assertEqual(
            minimist.parse('a b c'),
            dict(_=['a', 'b', 'c']))

    def test_abc_2(self):
        self.assertEqual(
            minimist.parse('-a -b -c'),
            dict(_=[], a=True, b=True, c=True))

    def test_abc_3(self):
        self.assertEqual(
            minimist.parse('-abc'),
            dict(_=[], a=True, b=True, c=True))

    def test_arg_eq_value(self):
        self.assertEqual(
            minimist.parse('--arg long-option -a short-option'),
            dict(_=[], arg='long-option', a='short-option'))

    def test_list(self):
        self.assertEqual(
            minimist.parse('-a a -b b -b', lists=['a']),
            dict(_=[], a=['a'], b=['b', True]))

    def test_number(self):
        self.assertEqual(
            minimist.parse('--pi 3.14 -r 10'),
            dict(_=[], pi=3.14, r=10))

    def test_strings(self):
        self.assertEqual(
            minimist.parse('--pi 3.14 -r 10', strings=['pi', 'r']),
            dict(_=[], pi='3.14', r='10'))

    def test_bool_arg(self):
        self.assertEqual(
            minimist.parse('--arg. -a.'),
            dict(_=[], arg=True, a=True))

    def test_more_value_1(self):
        self.assertEqual(
            minimist.parse('-a 1,2,3 -b 3e4'),
            dict(_=[], a=(1, 2, 3), b=3e4))

    def test_more_value_2(self):
        self.assertEqual(
            minimist.parse('-a "{1,2,3}"'),
            dict(_=[], a={1, 2, 3}))

    def test_defaults(self):
        self.assertEqual(
            minimist.parse('--pi 3.14 -r 10', defaults=dict(type='cycle', r=1)),
            dict(_=[], pi=3.14, r=10, type='cycle'))

    def test_extra_args(self):
        self.assertEqual(
            minimist.parse('a b c -- -d --e f'),
            dict(_=['a', 'b', 'c'], **{ '--':  ['-d', '--e', 'f']}))


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# -*- coding: utf-8 -*-
from aiohttp import web
from pu.util import reload_any, load_any


def reloadable_handler(real_handler, reload=True):
    '''自动重新加载的 handler

    :param real_handler: handler 对象或名称
    :param reload: 是否自动加载
    '''
    def loader(request):
        try:
            if reload:
                handler = reload_any(real_handler)
            else:
                handler = load_any(real_handler)
        except Exception as e:
            return web.Response(text='加载/重新加载 {} 时产生异常: {}'.format(real_handler, e))

        return handler(request)
    return loader



# -*- coding: utf-8 -*-
from functools import partial
from pu.dictutil import repr_dict, DotDict, DotOrderedDict, Dot, OrderedDict
import unittest


class Test(unittest.TestCase):
    def test_repr_dict(self):
        d = dict(a=4, b=3, c=2, d=1)
        repr_dict(d)
        self.assertEqual(repr_dict(d, sorted), 'dict(a=4, b=3, c=2, d=1)')
        self.assertEqual(repr_dict(d, partial(sorted, reverse=True)), 'dict(d=1, c=2, b=3, a=4)')

    def test_dotdict_getitem(self):
        d = DotDict(a=1)
        self.assertEqual(d.a, 1)

    def test_dotdict_setitem(self):
        d = DotDict()
        d.b = 2
        self.assertEqual(d['b'], 2)

    def test_dotdict_non_exists(self):
        d = DotDict()
        try:
            assert d.c
        except AttributeError:
            pass
    
    def test_dotordereddict(self):
        d = DotOrderedDict(a=1)
        d.b = 2
        assert d.a == 1
        assert d['b'] == 2
        assert list(d.keys()) == ['a', 'b']
        try:
            assert d.c
        except AttributeError:
            pass
    
    def test_dotit(self):
        adict = dict(a=1, b=2)
        d = Dot(adict)
        assert d.a == 1
        assert d['a'] == 1
        assert d.b == 2
        try:
            assert d.c
        except AttributeError:
            pass

    def test_dotit_nest_dict(self):
        adict = dict(a=dict(aa=1, ab=2), b=2)
        d = Dot(adict)
        assert d.a.aa == 1
        assert d.a.ab == 2

    def test_dotit_ordered_dict(self):
        adict = OrderedDict([
            ('a', OrderedDict([
                ('aa', 1),
                ('ab', 2),
                ])),
            ('b', 2),
            ])
        d = Dot(adict)
        assert d.a.aa == 1
        assert d.a.ab == 2

    def test_dotit_repr(self):
        adict = dict(a=dict(aa=1, ab=2), b=2)
        d = Dot(adict)
        self.assertEqual(repr(adict), repr(d))



# -*- coding: utf-8 -*-
from pu.util import make_key


class Cache(dict):
    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs):
        key = make_key(args, kwargs)
        try:
            return self[key]
        except KeyError:
            return self.setdefault(key, self.func(*args, **kwargs))



# -*- coding: utf-8 -*-
import asyncio


__all__ = ['Timer']


class Timer:
    __timer = None

    def __init__(self, callback=None, loop=None):
        self.callback = callback or self.timeout
        self.loop = loop or asyncio.get_event_loop()

    def set_timeout(self, delay):
        self.__timer = self.loop.call_later(delay, self._timeout)

    def cancel_timeout(self):
        if self.__timer:
            self.__timer.cancel()
            self.__timer = None

    def reset_timeout(self, delay):
        self.cancel_timeout()
        self.set_timeout(delay)

    def _timeout(self):
        self.__timer = None

        self.callback()

    def timeout(self):
        raise NotImplementedError


if __name__ == '__main__':
    import time
    start = time.time()

    now = lambda : round(time.time() - start, 6)

    loop = asyncio.get_event_loop()

    def timeout():
        print(now(), '时间到')
        loop.stop()

    def delay():
        print(now(), '推迟到 3 秒后')
        timer.reset_timeout(3)

    timer = Timer(timeout)
    print(now(), '启动 3 秒定时')
    timer.set_timeout(3)
    loop.call_later(2, delay)

    loop.run_forever()



# -*- coding: utf-8 -*-
'''Remote Call Protocol

远程过程调用协议，参考 JSON-RPC 定义，但针对 Python 语言特点做了如下改变:

- 简化: 保留必要的字段，用列表取代字典
- 明确: 明确定义了各种消息的格式

本协议的目标是提供一个最小的消息集合，至于扩展性等问题，可以通过上层协议进行处理。

四种消息类型:

Notify
------

通知消息，不需要对方回应，格式::

    ['N', 'method', args, kwargs]

- method:   字符串，方法名
- args:     列表或 tuple
- kwargs:   字典

Call
----

远程调用，对方返回 Return 或 Error 消息，格式::

    ['C', id, 'method', args, kwargs]

- id:       调用标识
- method:   方法名
- args:     列表或 tuple
- kwargs:   字典

Return
------

对 Call 的成功响应，格式::

    ['R', id, result]

- id:       对应 Call 中的 id
- result:   返回结果

Error
-----

错误响应， 格式::

    ['E', id, code, reason, data]

- id:       对应 Call 中的 id
- code:     错误代码
- reason:   错误描述
- data:     附加数据


常见错误代码::

- 101:      协议解析错误
- 102:      方法不存在
- 103:      参数错误
- 104:      内部错误
- 105~199:  其他错误


参考:

- http://en.wikipedia.org/wiki/JSON-RPC
- http://www.jsonrpc.org/specification
'''
from collections import namedtuple


__all__ = ['Notify', 'Call', 'Return', 'Error', 'decode', 'encode']


NOTIFY = 'N'
CALL = 'C'
RETURN = 'R'
ERROR = 'E'


class EncodeError(Exception):
    pass


class DecodeError(Exception):
    pass


Notify = namedtuple('Notify', ['method', 'args', 'kwargs'])
Call = namedtuple('Call', ['id', 'method', 'args', 'kwargs'])
Return = namedtuple('Return', ['id', 'result'])
Error = namedtuple('Error', ['id', 'code', 'reason', 'data'])


def decode(lst):
    message_type, *args = lst

    if message_type == NOTIFY:
        message = Notify(*args)
    elif message_type == CALL:
        message = Call(*args)
    elif message_type == RETURN:
        message = Return(*args)
    elif message_type == ERROR:
        message = Error(*args)
    else:
        raise DecodeError('解码失败: %r' % lst)

    return message


def encode(message):
    if isinstance(message, Notify):
        lst = [NOTIFY]
    elif isinstance(message, Call):
        lst = [CALL]
    elif isinstance(message, Return):
        lst = [RETURN]
    elif isinstance(message, Error):
        lst = [ERROR]
    else:
        raise EncodeError('编码失败: %r' % message)

    lst.extend(message)

    return lst



# -*- coding: utf-8 -*-
class Event(list):
    """事件订阅

    参考:

    - http://stackoverflow.com/questions/1092531/event-system-in-python/2022629#2022629
    """
    def __call__(self, *args, **kwargs):
        return [func(*args, **kwargs) for func in self]


class EventNamespace(dict):
    def event(self, name):
        if name in self:
            return self[name]
        return self.setdefault(name, Event())


event = EventNamespace().event



# -*- coding: utf-8 -*-
'''事件订阅管理

功能：

- 频道支持通配符: * ?
- 支持优先级，数值越低优先级越高，优先级可以为负数，缺省优先级为 0
- 允许终止后续消息处理: raise PubSub.Stop
'''
from collections import defaultdict
import fnmatch
import logging
from operator import itemgetter
import re


logger = logging.getLogger(__name__)


class PubSub(object):
    class Stop(Exception):
        pass

    def __init__(self):
        # 保存订阅关系
        self.channel_handlers = defaultdict(list)
        self.pattern_handlers = defaultdict(list)

        # 保存 handler 到 channel 的反向引用，用在取消订阅的时候
        self.handler_channels = defaultdict(list)
        self.handler_patterns = defaultdict(list)

    def subscribe(self, handler, *channels, priority=0):
        handler_info = (handler, priority)

        logger.info('subscribe({}, {})'.format(handler_info, channels))

        for channel in channels:
            self.channel_handlers[channel].append(handler_info)
            self.handler_channels[handler_info].append(channel)

    def psubscribe(self, handler, *patterns, priority=0):
        handler_info = (handler, priority)

        logger.info('psubscribe({}, {})'.format(handler_info, patterns))

        for pattern in patterns:
            pattern = re.compile(fnmatch.translate(pattern))
            self.pattern_handlers[pattern].append(handler_info)
            self.handler_patterns[handler_info].append(pattern)

    def unsubscribe(self, handler, *channels, priority=0):
        '''取消订阅，channels 为空时表示取消所有 handler 的订阅
        '''
        handler_info = (handler, priority)

        if not channels:
            channels = self.handler_channels[handler_info]

        logger.info('unsubscribe({}, {})'.format(handler_info, channels))

        for channel in channels:
            try:
                self.channel_handlers[channel].remove(handler_info)
                self.handler_channels[handler_info].remove(channel)
            except KeyError:
                logger.warning('取消事件注册失败: {} {}'.format(channel, handler_info))
            else:
                # 清理空的列表项
                if not self.channel_handlers[channel]:
                    del self.channel_handlers[channel]

        # 清理空的列表项，以免 handler 被占用导致资源泄漏
        if not self.handler_channels[handler_info]:
            del self.handler_channels[handler_info]

    def punsubscribe(self, handler, *patterns, priority=0):
        '''取消订阅，patterns 为空时表示取消所有 handler 的订阅
        '''
        handler_info = (handler, priority)

        if not patterns:
            patterns = self.handler_patterns[handler_info]

        logger.info('punsubscribe({}, {})'.format(handler_info, patterns))

        for pattern in patterns:
            try:
                self.pattern_handlers[pattern].remove(handler_info)
                self.handler_patterns[handler_info].remove(pattern)
            except KeyError:
                logger.warning('取消事件注册失败: {} {}'.format(pattern, handler_info))
            else:
                # 清理空的列表项
                if not self.pattern_handlers[pattern]:
                    del self.pattern_handlers[pattern]

        # 清理空的列表项，以免 handler 被占用导致资源泄漏
        if not self.handler_patterns[handler_info]:
            del self.handler_patterns[handler_info]

    def publish(self, channel, message, sender=None):
        handler_infos = self.channel_handlers[channel].copy()

        pattern_handlers = self.pattern_handlers
        for pattern in pattern_handlers:
            if pattern.match(channel):
                handler_infos.extend(pattern_handlers[pattern])

        # 按优先级排序
        handler_infos = sorted(handler_infos, key=itemgetter(1))
        handlers = (handler for handler, _ in handler_infos)

        count = 0

        for handler in handlers:
            try:
                handler(channel, message, sender)
            except PubSub.Stop:
                break
            except Exception as e:
                logger.exception('进行消息处理 {}({}, {}, {}) 时产生异常: {}'.format(
                    handler, channel, message, sender, e))
            else:
                count += 1

        return count


pubsub = PubSub()


if __name__ == '__main__':
    from functools import partial

    pubsub.subscribe(partial(print, 'subscribe'), 'comp.learn', 'comp.python', priority=-9)
    pubsub.psubscribe(partial(print, 'psubscribe'), 'comp.*')

    pubsub.publish('comp.python', 'Tom', 'I like python')
    pubsub.publish('comp.c', 'Bob', 'I like c')



# -*- coding: utf-8 -*-
from collections import OrderedDict


class Manager(OrderedDict):
    '''对象管理器

    用法::

        manager = Manager()

        manager.register(1, 'ONE')

        @manager.register
        def a(): pass

        @manager.named('funcb', 'FUNCB')
        def b(): pass

        @manager.register
        class A: pass

        @manager.named('clsb', 'CLSB')
        class B: pass
    '''
    def register(self, ob, name=None):
        if not name:
            name = ob.__name__
        assert name not in self
        self[name] = ob
        return ob

    def named(self, *names):
        def func(ob):
            for name in names:
                self[name] = ob
            return ob
        return func



# -*- coding: utf-8 -*-
from collections import Mapping
import importlib
import inspect
import re
import sys
import time


def shorten(s, width=80):
    '''
    >>> shorten('a very very very very long sentence', 20)
    'a very very ..(23)..'
    '''
    if not isinstance(s, str):
        s = str(s)

    length = len(s)
    if length < width:
        return s

    cut_length = length - width + 6
    x = len(str(cut_length))
    cut_length += x

    # 长度调整
    if x != len(str(cut_length)):
        cut_length += 1

    end_pos = length - cut_length
    return s[:end_pos] + '..(%d)..' % cut_length


def bytes_fromhex(s):
    return bytes.fromhex(re.sub('\s', '', s))


def iterattrs(ob, protect_member=False):
    '''返回对象的属性列表
    '''
    if protect_member:
        names = (n for n in sorted(dir(ob)) if not n.startswith('__'))
    else:
        names = (n for n in sorted(dir(ob)) if not n.startswith('_'))

    return iter((n, getattr(ob, n)) for n in names)


def deep_encode(ob, encoding='utf_8', errors='strict'):
    '''深入数据结构内部，尽可能把字符串编码
    '''
    if isinstance(ob, bytes):
        return ob
    elif isinstance(ob, str):
        return ob.encode(encoding, errors)
    elif isinstance(ob, tuple):
        return tuple(deep_encode(x, encoding, errors) for x in ob)
    elif isinstance(ob, list):
        return [deep_encode(x, encoding, errors) for x in ob]
    elif isinstance(ob, Mapping):
        new = ob.__class__()
        for key, value in ob.items():
            key = deep_encode(key, encoding, errors)
            value = deep_encode(value, encoding, errors)
            new[key] = value
        return new
    else:
        return ob


def deep_decode(ob, encoding='utf_8', errors='strict'):
    '''深入数据结构内部，尽可能把 bytes 解码
    '''
    if isinstance(ob, bytes):
        return ob.decode(encoding, errors)
    elif isinstance(ob, str):
        return ob
    elif isinstance(ob, tuple):
        return tuple(deep_decode(x, encoding, errors) for x in ob)
    elif isinstance(ob, list):
        return [deep_decode(x, encoding, errors) for x in ob]
    elif isinstance(ob, Mapping):
        new = ob.__class__()
        for key, value in ob.items():
            key = deep_decode(key, encoding, errors)
            value = deep_decode(value, encoding, errors)
            new[key] = value
        return new
    else:
        return ob


def format_args(args, kwargs):
    items = list(map(repr, args))
    if isinstance(kwargs, dict):
        kwargs = sorted(kwargs.items())
    items.extend(map(lambda kv: '%s=%r' % kv, kwargs))
    return ', '.join(items)


def format_time(t=None, fmt='%Y-%m-%d %H:%M:%S'):
    return time.strftime(fmt, time.localtime(t))


def get_field(o, field_name):
    '''支持多级对象的读取

    语法: .attr:index#index ...

    - .attr
    - :index     [index]
    - #index     [int(index)]
    '''
    # 开头的 . 可以省略
    if field_name[0] not in '.:#':
        field_name = '.' + field_name

    names = re.findall(r'[.:#]\w+', field_name)
    for name in names:
        flag = name[0]
        name = name[1:]
        if flag == '.':
            o = getattr(o, name)
        elif flag == ':':
            o = o[name]
        else:
            o = o[int(name)]

    return o


def set_field(o, field_name, value):
    '''多级对象属性的修改
    '''
    # 开头的 . 可以省略
    if field_name[0] not in '.:#':
        field_name = '.' + field_name

    * names, last_name = re.findall(r'[.:#]\w+', field_name)
    for name in names:
        flag = name[0]
        name = name[1:]
        if flag == '.':
            o = getattr(o, name)
        elif flag == ':':
            o = o[name]
        else:
            o = o[int(name)]

    flag, name = last_name[0], last_name[1:]
    if flag == '.':
        setattr(o, name, value)
    elif flag == ':':
        o[name] = value
    else:
        o[int(name)] = value


def to_bool(value):
    if isinstance(value, bool):
        return value
    elif isinstance(value, str):
        orig_value = value
        value = orig_value.strip().lower()

        if value in ['y', 'yes', 't', 'true', '1']:
            return True
        elif value in ['n', 'no', 'f', 'false', '0']:
            return False
        else:
            raise ValueError('%r 不能转换成布尔类型' % orig_value)
    else:
        return bool(value)


def to_hex(value: bytes, delimiter=' ', lower=True):
    fmt = '%02x' if lower else '%02X'
    return delimiter.join(fmt % b for b in value)


def import_file(module_name, filepath):
    import importlib.machinery

    loader = importlib.machinery.SourceFileLoader(module_name, filepath)
    return loader.load_module()


def load_any(name):
    '''根据对象名称加载对象
    '''
    try:
        return importlib.import_module(name)
    except ImportError:
        module_name, name = name.rsplit('.', 1)
        module = importlib.import_module(module_name)
        return getattr(module, name)


def _reload_module_by_name(module_name):
    module = sys.modules.get(module_name)

    if module:
        module = importlib.reload(module)
    else:
        module = importlib.import_module(module_name)

    return module


def reload_any(ob):
    '''reload 任何对象

    例如::

        reload_any(os.path)    # 直接重新加载对象
        reload_any('sys.path') # 根据名称加载对象

    :param ob: 对象(适合于有 __name__ 的对象)或对象名称
    '''
    # 对于模块，直接加载
    if inspect.ismodule(ob):
        return importlib.reload(ob)

    if isinstance(ob, str):
        name = ob

        # 作为模块名尝试一下
        try:
            return _reload_module_by_name(name)
        except ImportError:
            pass

        # 一定是 "模块.对象" 的形式
        assert '.' in name
        module_name, ob_name = ob.rsplit('.', 1)
    else:
        module_name = ob.__module__
        ob_name = ob.__name__

    module = _reload_module_by_name(module_name)

    return getattr(module, ob_name)


def make_key(args, kwargs, kwargs_mark=object()):
    '''根据函数参数 args, kwargs 创建键值

    特点: 生成一个没有嵌套（args 和 kwargs 的值中没有 tuple 值）的 tuple
    限制: args 和 kwargs 的值必须 hashable

    :param args: tuple
    :param kwargs: dict
    :param kwargs_mark: 在有 kwargs 非空的情况下，用来分割 args 和 kwargs
    '''
    key = args

    if kwargs:
        key += kwargs_mark
        for item in sorted(kwargs.items):
            key += item

    return key


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
import binascii


class pretty_bytes(bytes):
    r'''提供自定义格式

    >>> pb = pretty_bytes(b'\xaa\xbb\xcc\xdd\xee\xff')
    >>> '{0:hex}'.format(pb)
    'aabbccddeeff'
    >>> '{0:HEX}'.format(pb)
    'AABBCCDDEEFF'
    >>> '{0:hex+}'.format(pb)
    'aa bb cc dd ee ff'
    >>> '{0:HEX+}'.format(pb)
    'AA BB CC DD EE FF'
    >>> '{0}'.format(pb)
    "b'\\xaa\\xbb\\xcc\\xdd\\xee\\xff'"
    '''
    def __format__(self, format_spec):
        if format_spec == 'hex':
            return binascii.hexlify(self).decode()
        elif format_spec == 'HEX':
            return ''.join('%02X' % b for b in self)
        elif format_spec == 'hex+':
            return ' '.join('%02x' % b for b in self)
        elif format_spec == 'HEX+':
            return ' '.join('%02X' % b for b in self)
        else:
            return str(self)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-
import asyncio


@asyncio.coroutine
def file_get_contents(filename, loop=None):
    '''读取文件内容，支持本地文件和 http 链接

    这里只提供最简单接口，如果需要更复杂的控制，请直接调用相应函数
    '''
    if filename.startswith('http://'):
        import aiohttp      # 避免不必要的依赖

        if not loop:
            loop = asyncio.get_event_loop()

        response = yield from aiohttp.request('GET', filename)
        content = yield from response.read()
        if response.status != 200:
            raise RuntimeError('读取文件内容失败: %d, %s, %r' % (
                response.status, filename, content))
        return content
    else:
        with open(filename, 'rb') as f:
            return f.read()



# -*- coding: utf-8 -*-
import asyncio
import logging
import textwrap

from pu import minimist
from pu.aio.protocols.basic import LineReceiver


logger = logging.getLogger(__name__)


class Error(Exception):
    code = 100
    message = '错误'

    def __init__(self, message=None):
        super().__init__(message)

        if message:
            self.message = message


class DecodeError(Error):
    code = 101
    message = '解码错误'


class CommandNotFound(Error):
    code = 102
    message = '命令不存在'


class BadFormat(Error):
    code = 103
    message = '命令格式错误'


class InvalidParam(Error):
    code = 104
    message = '无效参数'


class Cli(LineReceiver):
    encoding = 'utf_8'

    def __init__(self, password=None):
        self.password = password

    def write_line(self, line):
        if not self.password:
            if isinstance(line, str):
                line = line.encode(self.encoding)
            super().write_line(line)

    def write(self, data):
        if not self.password:
            self.transport.write(data)

    def _write_lines(self, lines, prefix='M'):
        if self.password:
            return
        if isinstance(lines, str):
            lines = [lines]
        self.write_line(prefix + ' ' + lines.pop(0))
        for line in lines:
            self.write_line('  ' + line)

    def write_error(self, message):
        self._write_lines(message, 'E')

    def write_response(self, message):
        self._write_lines(message, '-')

    def write_notify(self, message):
        self._write_lines(message, 'N')

    def splitlines(self, s):
        return s.split(self._delimiter.decode())

    def line_received(self, line):
        try:
            line = line.decode(self.encoding).strip()
        except Exception as e:
            raise DecodeError()

        if not line:
            return

        if line.startswith('#'):
            return

        kwargs = minimist.parse(line, comments=False)

        command, *args = kwargs.pop('_')

        handler_name = command.upper()

        try:
            handler = self.get_handler(handler_name)
        except Error as e:
            self.write_error('%d: %s' % (e.code, e.message))
        else:
            if not self.password or handler.__name__ == 'AUTH':
                try:
                    handler(*args, **kwargs)
                except Error as e:
                    self.write_error('%d: %s' % (e.code, e.message))
                except Exception as e:
                    logger.exception('处理 %s 出错: %s', handler.__name__, e)
                    self.write_error('999: %s' % (e))

    def get_handler(self, handler_name):
        handler = getattr(self, handler_name, None)

        if not handler:
            # 支持命令缩写
            handler_names = [n for n in dir(self) if n.startswith(handler_name)]
            if len(handler_names) != 1:
                raise CommandNotFound()
            handler_name = handler_names[0]
            handler = getattr(self, handler_name)

        return handler

    def AUTH(self, password):
        if self.password == password:
            self.password = None

    def HELP(self, *commands):
        '''查看帮助信息
        '''
        docs = []
        if commands:
            for command in commands:
                handler = getattr(self, command.upper(), None)
                if handler:
                    # 重新格式化文档
                    doc = handler.__doc__ or '-'
                    lines = doc.splitlines()

                    docs.append('%-20s%s' % (command, lines.pop(0)))

                    # 处理后续行的缩进问题
                    lines = textwrap.dedent('\n'.join(lines)).splitlines()

                    for line in lines:
                        docs.append('%-20s%s' % (' ', line))
                else:
                    docs.append('%-20s不存在' % command)
        else:
            commands = [m for m in dir(self) if m.upper() == m]
            for command in commands:
                handler = getattr(self, command.upper(), None)
                doc = handler.__doc__ or '-'
                doc = doc.splitlines()[0]
                doc = '%-20s%s' % (command, doc)
                docs.append(doc)

        self.write_response(docs)

    def set_encoding(self, encoding):
        try:
            assert '测试'.encode(encoding).decode(encoding) == '测试'
        except:
            raise InvalidParam('无效编码类型 %s' % encoding)
        else:
            self.encoding = encoding

    def ENCODING(self, encoding=None):
        '''设置编码
        命令格式: encoding [gbk|utf-8|...]
        '''
        if encoding:
            if self.encoding != encoding:
                self.set_encoding(encoding)
        else:
            self.write_response('encoding: %s' % self.encoding)

    def GBK(self):
        self.ENCODING('gbk')

    def UTF8(self):
        self.ENCODING('utf_8')


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(loop.create_server(
        lambda : Cli('aa'), '0.0.0.0', 9999))
    loop.run_forever()



# -*- coding: utf-8 -*-
"""Load YAML file

Feature:

- support !include tag to include another YAML file
"""
from __future__ import print_function
from yaml.constructor import Constructor, ConstructorError
import os.path
import yaml


INCLUDE_TAG = '!include'


def load(filename, encoding='utf-8'):
    dirname = os.path.dirname(filename)

    def include(loader, node):
        try:
            filename = loader.construct_scalar(node)
        except:
            try:
                filename, nodename = loader.construct_sequence(node)
            except:
                raise ConstructorError(None, None,
                    'expected a string or sequence(2 elements), but found %s' % node.id,
                    node.start_mark)
        else:
            nodename = None

        with open(os.path.join(dirname, filename), encoding=encoding) as f:
            document = yaml.load(f)

        if nodename:
            return document[nodename]

        return document

    Constructor.add_constructor(INCLUDE_TAG, include)

    with open(filename, encoding=encoding) as f:
        data = yaml.load(f)

    del Constructor.yaml_constructors[INCLUDE_TAG]

    return data



# -*- coding: utf-8 -*-
import unittest
from urllib import parse

from pu.url import parse_url, Netloc


class TestParseUrl(unittest.TestCase):
    def test_host_only(self):
        result = parse_url('hostname')
        self.assertEqual(result.hostname, 'hostname')
        self.assertEqual(result.port, None)

    def test_port_only(self):
        result = parse_url('9999', default_host='localhost')
        self.assertEqual(result.hostname, 'localhost')
        self.assertEqual(result.port, 9999)

    def test_hostport(self):
        result = parse_url('hostname:9999')
        self.assertEqual(result.hostname, 'hostname')
        self.assertEqual(result.port, 9999)

    def test_path_query(self):
        result = parse_url('hostname/path?x=1&y=2')
        self.assertEqual(parse.parse_qsl(result.query), [('x', '1'), ('y', '2')])

    def test_params(self):
        result = parse_url('hostname/path;params is here')
        self.assertEqual(result.path, '/path;params is here')
        self.assertEqual(result.params, '')
        result = parse_url('hostname/path;params is here', parse_params=True)
        self.assertEqual(result.params, 'params is here')

    def test_misc(self):
        result = parse_url('http://www.baidu.com/s?ie=utf-8&f=8&rsv_bp=1&ch=&tn=baidu&bar=&wd=parse_url')
        self.assertIn(('wd', 'parse_url'), parse.parse_qsl(result.query))


class TestNetloc(unittest.TestCase):
    def test_netloc(self):
        result = Netloc('9999', default_host='localhost')

        self.assertEqual(result.hostname, 'localhost')
        self.assertEqual(result.port, 9999)



# -*- coding: utf-8 -*-
import asyncio
import logging

logging.warning('%s 已经废弃，请使用 dummyprotocol.', __name__)

__all__ = ['VirtualProtocol', 'RealProtocol']

logger = logging.getLogger(__name__)


class VirtualProtocol(asyncio.Protocol):
    _real_protocol = None
    _protocol_detect_buffer = b''

    def __init__(self, factory):
        self._protocol_factories = factory._protocol_factories.copy()
        self._default_protocol_factory = factory._default_protocol_factory

    def set_real_protocol(self, protocol):
        self._real_protocol = protocol
        self._real_protocol.connection_made(self.transport)
        self.data_received = protocol.data_received
        self.connection_lost = protocol.connection_lost

    def connection_made(self, transport):
        self.transport = transport
        self.peername = '%s:%d' % transport.get_extra_info('peername')
        self.sockname = '%s:%d' % transport.get_extra_info('sockname')

        logger.debug('连接建立: 远端 %s, 本地 %s', self.peername, self.sockname)

    def connection_lost(self, exc):
        assert not self._real_protocol

        logger.debug('连接断开: %s, 远端 %s, 本地 %s', exc, self.peername, self.sockname)

    def data_received(self, data):
        assert not self._real_protocol

        data = self._protocol_detect_buffer + data

        for factory in self._protocol_factories[:]:
            result = factory.protocol_detect(data)
            if result == factory.YES:
                logger.debug('协议检测成功: %s 采用 %s 协议', self.peername, factory.__name__)
                self.set_real_protocol(factory(self))
                break
            elif result == factory.NO:
                self._protocol_factories.remove(factory)

        if not self._protocol_factories:
            if self._default_protocol_factory:
                factory = self._default_protocol_factory
                logger.info('%s 采用缺省协议: %s', self.peername, factory.__name__)
                self.set_real_protocol(factory(self))
            else:
                logger.info('不可识别的协议, 断开连接 %s: %r', self.peername, data)
                self.transport.close()
        elif self._real_protocol:
            self._real_protocol.data_received(data)
        else:
            self._protocol_detect_buffer = data


class VirtualProtocolFactory:
    def __init__(self, virtual_protocol_class=VirtualProtocol):
        self.virtual_protocol_class = virtual_protocol_class

        self._protocol_factories = []
        self._default_protocol_factory = None

    def register_protocol_factory(self, protocol_factory, default=False):
        if default:
            assert not self._default_protocol_factory
            self._default_protocol_factory = protocol_factory
        else:
            assert protocol_factory not in self._protocol_factories
            self._protocol_factories.append(protocol_factory)

    def __call__(self):
        return self.virtual_protocol_class(self)


class RealProtocol:
    NO = 'NO'
    YES = 'YES'
    NOTSURE = 'NOTSURE'

    protocol_features = []      # bytes 列表

    @classmethod
    def protocol_detect(cls, data):
        if not cls.protocol_features:
            raise RuntimeError('未指定协议特征，无法检测.')

        notsure = False
        for feature in cls.protocol_features:
            if data.startswith(feature):
                return cls.YES
            if feature.startswith(data):
                # 如果 feature 以 data 开头，说明数据不够，不足以判别
                notsure = True

        if notsure:
            return cls.NOTSURE

        return cls.NO

    def __init__(self, virtual_protocol):
        self.virtual_protocol = virtual_protocol

    def connection_made(self, transport):
        self.transport = transport
        self.peername = self.virtual_protocol.peername
        self.sockname = self.virtual_protocol.sockname

    def connection_lost(self, exc):
        logger.debug('连接断开: %s, 远端 %s, 本地 %s', exc, self.peername, self.sockname)

    def data_received(self, data):
        logger.debug('来自 %s : %r', self.peername, data)


if __name__ == '__main__':
    class PYES(RealProtocol):
        protocol_features = [b'YES']

    class PNO(RealProtocol):
        protocol_features = [b'NO']

    class PNOTSURE(RealProtocol):
        protocol_features = [b'NOTSURE']

    class PXX(RealProtocol):
        @classmethod
        def protocol_detect(cls, data):
            if len(data) < 2:
                return cls.NOTSURE
            if data[0] == data[1]:
                return cls.YES
            return cls.NO

    logging.basicConfig(level='DEBUG')

    factory = VirtualProtocolFactory()
    factory.register_protocol_factory(PYES)
    factory.register_protocol_factory(PNOTSURE)
    factory.register_protocol_factory(PNO, True)
    factory.register_protocol_factory(PXX)

    loop = asyncio.get_event_loop()
    host, port = '0.0.0.0', 9999
    loop.run_until_complete(loop.create_server(factory, host, port))
    logger.info('监听: %s:%d', host, port)
    loop.run_forever()



# -*- coding: utf-8 -*-
from urllib import parse


def parse_url(url, default_host=None, parse_params=False):
    '''更灵活的 url 分析

    url 格式:

        <scheme>://<netloc>/<path>;<params>?<query>#<fragment>

    netloc 格式(未指定 default_host 时):

        [<username>[:<password>]]@<hostname>[:<port>]

    netloc 格式(指定了 default_host 时):

        [<username>[:<password>]]@[<hostname>:]<port>

    :param parse_params: 对于未知协议，指定是否分析 params 字段
    '''
    no_scheme = '://' not in url

    if no_scheme:
        if parse_params:
            scheme_na = parse.uses_params[0]
        else:
            scheme_na = 'unknown'

        url = scheme_na + '://' + url

    result = parse.urlparse(url)
    scheme, netloc, path, params, query, fragment = result

    netloc = Netloc(netloc, default_host).netloc

    if no_scheme:
        scheme = ''

    return parse.ParseResult(scheme, netloc, path, params, query, fragment)


def parse_hostport(hostport, default_host=None):
    return Netloc(hostport, default_host).hostinfo


class Netloc(parse._NetlocResultMixinStr):
    def __init__(self, netloc, default_host=None):
        self.netloc = netloc

        # 处理 hostname:port 不完整的情况
        if self.port is None and default_host:
            self.netloc = '{}:{}'.format(default_host, netloc)

    def __str__(self):
        return self.netloc

    @property
    def hostinfo(self):
        return self.hostname, self.port

    @property
    def userinfo(self):
        return self.username, self.password


class URL(parse._NetlocResultMixinStr):
    def __init__(self, url, default_host=None):
        self.url = url
        self.default_host = default_host

        self.parse()

    def parse(self):
        result = parse_url(self.url, self.default_host)

        self.netloc = result.netloc

        return result

    def __str__(self):
        return self.url


class MySQL(URL):
    db = None
    charset = 'utf8'

    def parse(self):
        result = URL.parse(self)

        for name, value in parse.parse_qsl(result.query):
            if name == 'charset':
                self.charset = value
            else:
                raise Warning('未知参数: {}={}'.format(name, value))

        if result.path:
            self.db = result.path.lstrip('/')



# -*- coding: utf-8 -*-
class Observable:
    '''观察者模式

    为了，简化接口，约定观察者为函数。

    用法:

        observable = Observable()
        observable.add_observer(print)
        observable.notify_observers('hello')
    '''
    _observers = None

    def add_observer(self, observer):
        '''添加观察者（可调用）

        :param observer: 观察者
        '''
        assert callable(observer)

        if self._observers is None:
            self._observers = []

        self._observers.append(observer)

    def remove_observer(self, observer):
        '''移除观察者

        :param observer: 观察者
        '''
        assert observer in self._observers

        self._observers.remove(observer)

    def notify_observers(self, *args, **kwargs):
        '''通知观察者

        :param message: 消息
        '''
        for observer in self._observers:
            observer(*args, **kwargs)


# Observerable 是拼写错误，为了兼容原来的程序，保留这个名称
Observerable = Observable



# -*- coding: utf-8 -*-
from pu.simplefilter import Filter
import unittest


class FitlerTest(unittest.TestCase):
    def test_equal(self):
        f = Filter('a = A')

        self.assertTrue(f(dict(a='A')))

        self.assertFalse(f(dict()))
        self.assertFalse(f(dict(a='a')))

    def test_not_equal(self):
        f = Filter('a != A')

        self.assertTrue(f(dict(a='a')))

        self.assertTrue(f(dict()))
        self.assertFalse(f(dict(a='A')))

    def test_almost_equal(self):
        f = Filter('a ~= *A*')

        self.assertTrue(f(dict(a='A')))
        self.assertTrue(f(dict(a='--A--')))

        self.assertFalse(f(dict()))
        self.assertFalse(f(dict(a='a')))

    def test_not_almost_equal(self):
        f = Filter('a !~= *A*')

        self.assertFalse(f(dict(a='A')))
        self.assertFalse(f(dict(a='--A--')))

        self.assertTrue(f(dict()))
        self.assertTrue(f(dict(a='a')))

    def test_hex_equal(self):
        f = Filter('a #= AA')

        self.assertTrue(f(dict(a=b'\xAA')))

        self.assertFalse(f(dict()))
        self.assertFalse(f(dict(a=b'a')))

    def test_hex_not_equal(self):
        f = Filter('a #!= AA')

        self.assertTrue(f(dict(a='a')))

        self.assertTrue(f(dict()))
        self.assertFalse(f(dict(a='\xAA')))

    def test_hex_almost_equal(self):
        f = Filter('a #~= *AA*')

        self.assertTrue(f(dict(a='\xAA')))
        self.assertTrue(f(dict(a='--\xAA--')))

        self.assertFalse(f(dict()))
        self.assertFalse(f(dict(a='a')))

    def test_hex_not_almost_equal(self):
        f = Filter('a #!~= *AA*')

        self.assertFalse(f(dict(a='\xAA')))
        self.assertFalse(f(dict(a='--\xAA--')))

        self.assertTrue(f(dict()))
        self.assertTrue(f(dict(a='a')))


# TODO:
class AndFilterGroupTest(unittest.TestCase):
    pass


# TODO:
class OrFilterGroupTest(unittest.TestCase):
    pass


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import asyncio
import functools
import logging


__all__ = [
    'YES', 'NO', 'UNSURE',
    'DummyProtocolFactory',
    ]

YES = 'YES'
NO = 'NO'
UNSURE = 'UNSURE'

logger = logging.getLogger('dummyprotocol')


class DummyProtocol(asyncio.Protocol):
    _real_protocol = None
    _protocol_detect_buffer = b''

    def __init__(self, factory):
        self._protocol_factories = factory._protocol_factories.copy()
        self._default_protocol_factory = factory._default_protocol_factory

    def set_real_protocol(self, protocol):
        # 重定向自身的方法
        self.data_received = protocol.data_received
        self.eof_received = protocol.eof_received
        self.connection_lost = protocol.connection_lost

        # 通知 real protocol 连接建立
        protocol.connection_made(self.transport)

        # 数据移交给真正的协议对象
        protocol.data_received(self._protocol_detect_buffer)

        self._real_protocol = protocol

        # 清理无用数据和引用
        del self._protocol_factories
        del self._default_protocol_factory
        del self._protocol_detect_buffer
        del self.transport

    def connection_made(self, transport):
        self.transport = transport
        self.peername = '%s:%d' % transport.get_extra_info('peername')
        self.sockname = '%s:%d' % transport.get_extra_info('sockname')

        logger.debug('连接建立: 远端 %s, 本地 %s', self.peername, self.sockname)

    def connection_lost(self, exc):
        assert not self._real_protocol

        logger.debug('连接断开: %s, 远端 %s, 本地 %s', exc, self.peername, self.sockname)

    def data_received(self, data):
        assert not self._real_protocol

        buffer = self._protocol_detect_buffer + data
        self._protocol_detect_buffer = buffer

        # 协议检测
        for item in self._protocol_factories[:]:
            factory, detector = item

            result = detector(buffer)
            if result == YES:
                logger.debug('协议检测成功: %s 采用 %s 协议', self.peername, factory.__name__)
                self.set_real_protocol(factory())
                return
            elif result == NO:
                self._protocol_factories.remove(item)

        # 检查是否有缺省协议
        if not self._protocol_factories:
            if self._default_protocol_factory:
                factory = self._default_protocol_factory
                logger.info('%s 采用缺省协议: %s', self.peername, factory.__name__)
                self.set_real_protocol(factory())
            else:
                logger.info('不可识别的协议, 断开连接 %s: %r', self.peername, buffer)
                self.transport.close()


def _feature_detector(data, features):
    '''特征串检测函数

    :param data: 数据
    :param features: 特征串
    '''
    unsure = False

    if isinstance(features, bytes):
        features = [features]

    for feature in features:
        if data.startswith(feature):
            return YES
        if feature.startswith(data):
            # 如果 feature 以 data 开头，说明数据不够，不足以判别
            unsure = True

    return UNSURE if unsure else NO


class DummyProtocolFactory:
    def __init__(self, dummy_protocol_class=DummyProtocol):
        self.dummy_protocol_class = dummy_protocol_class

        self._protocol_factories = []
        self._default_protocol_factory = None

    def register_protocol_factory(self, protocol_factory, *, detector=None, features=None, default=None):
        '''注册协议，同时指定检测函数或特征串

        - 检测函数: detector(data), 缺省为 protocol_factory.detector
        - 特征串: bytes 或 [bytes, ...]，缺省为 protocol_factory.features
        - 缺省协议: 其他协议检测完毕，则不做进一步检测，直接采用该协议 

        :param protocol_factory: 要注册的协议
        :param detector: 检测函数
        :param features: 特征串
        :param default: 是否缺省协议
        '''
        assert not (detector and features)

        if not detector and not features:
            detector = getattr(protocol_factory, 'detector', None)
            if not detector:
                features = getattr(protocol_factory, 'features', None)

        if features:
            assert not detector
            detector = functools.partial(_feature_detector, features=features)

        if default is None:
            default = getattr(protocol_factory, 'default', False)

        if default:
            assert not self._default_protocol_factory
            self._default_protocol_factory = protocol_factory
        else:
            assert protocol_factory not in self._protocol_factories
            self._protocol_factories.append((protocol_factory, detector))

    def __call__(self):
        return self.dummy_protocol_class(self)


if __name__ == '__main__':
    import logging

    class BaseProtocol(asyncio.Protocol):
        def connection_made(self, transport):
            logging.info('%s, connection_made', self.__class__.__name__)

        def connection_lost(self, exc):
            logging.info('%s connection_lost: %s', self.__class__.__name__, exc)

        def data_received(self, data):
            logging.info('%s data_received: %s', self.__class__.__name__, data)

    class A(BaseProtocol):
        features = b'A'

    class B(BaseProtocol):
        features = [b'B', b'b']

    class C(BaseProtocol):
        pass

    def C_detector(data):
        if data.startswith(b'C'):
            return YES
        return NO

    class D(BaseProtocol):
        pass

    class XX(BaseProtocol):
        @staticmethod
        def detector(data):
            if len(data) < 2:
                return UNSURE
            if data[0] == data[1]:
                return YES
            return NO

    logging.basicConfig(level='DEBUG')

    factory = DummyProtocolFactory()
    factory.register_protocol_factory(A)
    factory.register_protocol_factory(B)
    factory.register_protocol_factory(C, detector=C_detector)
    factory.register_protocol_factory(D, default=True)
    factory.register_protocol_factory(XX)

    loop = asyncio.get_event_loop()
    host, port = '0.0.0.0', 9999
    loop.run_until_complete(loop.create_server(factory, host, port))
    logger.info('监听: %s:%d', host, port)
    loop.run_forever()



# -*- coding: utf-8 -*-
import asyncio


class _PauseableMixin:
    _paused = False

    def pause_reading(self):
        '''暂停读取和处理已经读取的数据
        '''
        self._paused = True
        self.transport.pause_reading()

    def resume_reading(self):
        '''恢复处理
        '''
        self._paused = False
        self.transport.resume_reading()
        self.data_received(b'')


class LineReceiver(asyncio.Protocol, _PauseableMixin):
    '''每次接收一行，支持自动探测行结束符，支持暂停处理

    由于有些数据已经接收到缓冲区内，因此需要 _paused 标志协助处理
    '''
    _buffer = b''
    _delimiter = None

    def _guess_delimiter(self):
        if b'\r\n' in self._buffer:
            self._delimiter = b'\r\n'
        elif b'\n' in self._buffer:
            self._delimiter = b'\n'
        elif b'\r' in self._buffer:
            self._delimiter = b'\r'

    def connection_made(self, transport):
        self.transport = transport

    def write_line(self, line):
        delimiter = self._delimiter or b'\r\n'
        data = line + delimiter
        self.transport.write(data)

    def data_received(self, data):
        self._buffer += data

        if not self._delimiter:
            self._guess_delimiter()
            if not self._delimiter:
                return

        while self._buffer and not self._paused:
            try:
                line, self._buffer = self._buffer.split(self._delimiter, 1)
            except:
                break
            self.line_received(line)

    def line_received(self, line):
        raise NotImplementedError


class IntNStringReceiver(asyncio.Protocol, _PauseableMixin):
    _buffer = b''
    _prefix_length = 0

    def connection_made(self, transport):
        self.transport = transport

    def write_string(self, string):
        strlen = len(string)
        prefix = strlen.to_bytes(self._prefix_length, 'big')
        self.transport.write(prefix + string)

    def data_received(self, data):
        data = self._buffer + data
        prefix_length = self._prefix_length

        while len(data) > prefix_length and not self._paused:
            strlen = int.from_bytes(data[:prefix_length], 'big')
            if len(data) <prefix_length + strlen:
                break
            endpos = prefix_length + strlen
            string, data = data[prefix_length:endpos], data[endpos:]
            self.string_received(string)

        self._buffer = data

    def string_received(self, string):
        raise NotImplementedError


class Int8StringReceiver(IntNStringReceiver):
    _prefix_length = 1


class Int16StringReceiver(IntNStringReceiver):
    _prefix_length = 2


class Int32StringReceiver(IntNStringReceiver):
    _prefix_length = 4



# -*- coding: utf-8 -*-
import functools
import unittest

from pu import util


class Test(unittest.TestCase):
    def test_shorten(self):
        s = 'Hi, this is shorter than orignal message ...'
        self.assertEqual(len(util.shorten(s, 30)), 30)
        self.assertEqual(len(util.shorten(s, 10)), 10)

    def test_shorten_cn(self):
        s = '有时候中文的宽度是个问题，字符数、显示所占宽度，以及编码后的字节数……'
        self.assertEqual(len(util.shorten(s, 30)), 30)
        self.assertEqual(len(util.shorten(s, 10)), 10)

    def test_bytes_fromhex(self):
        s = 'aa\tbb\ncc\rdd ee'
        self.assertRaises(ValueError, functools.partial(bytes.fromhex, s))
        self.assertEqual(util.bytes_fromhex(s), b'\xaa\xbb\xcc\xdd\xee')

    def test_format_time(self):
        # 由于时区问题，下面的测试在 Travis CI 中出现错误
        # self.assertEqual(util.format_time(1416469808), '2014-11-20 15:50:08')
        pass


class IterattrsTest(unittest.TestCase):
    class C(object):
        pass

    def setUp(self):
        self.o = IterattrsTest.C()
        self.o.a = 'a'
        self.o._a = '_a'

    def test_iterattrs(self):
        result = tuple(util.iterattrs(self.o))
        self.assertEqual(result, (('a', 'a'),))

    def test_iterattrs_with_filter(self):
        result = tuple(util.iterattrs(self.o, True))
        self.assertEqual(result, (('_a', '_a'), ('a', 'a')))


class CodecTest(unittest.TestCase):
    def setUp(self):
        self.string = ([{'1': 1}],)
        self.binary = ([{b'1': 1}],)

    def test_deep_encode(self):
        self.assertEqual(util.deep_encode(self.string), self.binary)

    def test_deep_decode(self):
        self.assertEqual(util.deep_decode(self.binary), self.string)


class GetSetFieldTest(unittest.TestCase):
    def setUp(self):
        class O: pass
        self.o = O()
        self.o.d = {}
        self.o.d['o'] = O()
        self.o.d['o'].l = [0, 1]
        self.o.a = 'a'

    def test_get_field(self):
        self.assertEqual(util.get_field(self.o, '.a'), 'a')
        self.assertEqual(util.get_field(self.o, '.d:o.l#1'), 1)

    def test_set_field(self):
        util.set_field(self.o, '.a', 'A')
        self.assertEqual(util.get_field(self.o, '.a'), 'A')
        util.set_field(self.o, '.d:o.l#1', 'hello')
        self.assertEqual(util.get_field(self.o, '.d:o.l#1'), 'hello')


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import unittest

from pu.lazydict import lazydict


class TestLazydict(unittest.TestCase):
    def setUp(self):
        self.d = lazydict()

    def test_setdefault(self):
        self.d.setdefault('a', 1000)
        self.assertEqual(self.d['a'], 1000)

    def test_setdefault_lazy(self):
        self.d.setdefault('a', factory=lambda: 1000)
        self.assertEqual(self.d['a'], 1000)

    def test_smart_get(self):
        self.d['a'] = {'b': [lazydict]}
        lazydict_name = self.d.smart_get('a.b.0.__name__')
        self.assertEqual(lazydict_name, 'lazydict')

    def test_smart_get_default(self):
        self.d['a'] = {'b': [lazydict]}
        value = 'Default!'
        new_value = self.d.smart_get('a.b.0.xxxx', value)
        self.assertEqual(new_value, value)

    def test_smart_get_exception(self):
        self.d['a'] = {'b': [lazydict]}
        with self.assertRaises(ValueError):
            self.d.smart_get('a.b.xxxx')

    def test_smart_set_dict(self):
        self.d['a'] = {}
        value = 'Yeah!'
        self.d.smart_set('a.b', value)
        new_value = self.d.smart_get('a.b')
        self.assertEqual(new_value, value)

    def test_smart_set_attr(self):
        self.d['a'] = {'b': [lazydict]}
        value = 'Yeah!'
        self.d.smart_set('a.b.0.newattr', value)
        new_value = self.d.smart_get('a.b.0.newattr')
        self.assertEqual(new_value, value)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
'''
大道至简，推荐使用 Dict。
'''

from collections import OrderedDict as _OrderedDict, Mapping


def repr_dict(d, sort_func=None):
    '''更灵活地显示 dict

    repr_dict(d, sorted)
    repr_dict(d, partial(sorted, reverse=True))
    '''
    if d:
        if sort_func:
            keys = sort_func(d.keys())
        else:
            keys = d.keys()
        repr_items = ', '.join('%s=%r' % (k, d[k]) for k in keys)
    else:
        repr_items = ''
    return '%s(%s)' % (d.__class__.__name__, repr_items)


class Dot:
    '''支持对现有的 dict, OrderedDict 添加 . 方式访问
    '''
    def __init__(self, d):
        object.__setattr__(self, '_d', d)

    def __repr__(self):
        return self._d.__repr__()

    def __str__(self):
        return self._d.__str__()

    def __contains__(self, key):
        return self._d.__contains__(key)

    def __getitem__(self, key):
        return self._d.__getitem__(key)

    def __setitem__(self, key, value):
        return self._d.__setitem__(key, value)

    def __delitem__(self, key):
        return self._d.__delitem__(key)

    def __getattr__(self, name):
        try:
            value = self._d[name]
        except KeyError:
            raise AttributeError('%r has not attr %r' % (self, name))
        else:
            if isinstance(value, Mapping):
                return Dot(value)
            return value

    def __setattr__(self, name, value):
        return self._d.__setitem__(name, value)

    def __delattr__(self, name):
        try:
            del self._d[name]
        except KeyError:
            raise AttributeError('%r has not attr %r' % self, name)

    def get(self, key, default=None):
        return self._d.get(key, default)
    
    def setdefault(self, key, default):
        return self._d.setdefault(key, default)


class OrderedDict(_OrderedDict):
    '''改进显示，看起来更符合人们的习惯
    '''
    def __repr__(self):
        return repr_dict(self)


class DotDict(dict):
    '''允许用 . 访问字典成员
    '''
    __setattr__ = dict.__setitem__

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError('%r has not attr %r' % (self, name))

    def __delattr__(self, name):
        try:
            del self[name]
        except KeyError:
            raise AttributeError('%r has not attr %r' % (self, name))


class DotOrderedDict(OrderedDict):
    def __getattr__(self, name):
        if name.startswith('_OrderedDict'):
            raise AttributeError('%s has not attr %r' % (
                self.__class__.__name__, name))
        try:
            return self[name]
        except KeyError:
            raise AttributeError('%r has not attr %r' % (self, name))

    def __setattr__(self, name, value):
        if name.startswith('_OrderedDict'):
            return OrderedDict.__setattr__(self, name, value)
        self.__setitem__(name, value)

    def __delattr__(self, name):
        try:
            OrderedDict.__delitem__(self, name)
        except KeyError:
            raise AttributeError('%r has not attr %r' % (self, name))


class Dict(dict):
    '''增强的 dict

    - 转换成字符串时，按指定顺序显示字段，未指定顺序的按照 key 的排序
    '''
    _order_by = ()

    def __repr__(self):
        keys = [k for k in self._order_by if k in self]
        keys.extend(sorted(k for k in self if k not in self._order_by))
        items = ('{!r}: {!r}'.format(k, self[k]) for k in keys)
        return '{{{}}}'.format(', '.join(items))



# -*- coding: utf-8 -*-
import binascii
import fnmatch
import re


class Filter:
    def __init__(self, string=None):
        if string:
            self.parse(string)

    def __repr__(self):
        return '{name}{op}{pattern}'.format_map(self.__dict__)

    def parse(self, string):
        name, op, pattern = re.split('([#!~=]*)', string, 1)

        self.name = name.strip()
        self.op = op.strip()
        self.pattern = pattern.strip()

    def __call__(self, ob):
        '''过滤对象

        检查顺序:

        - 优先检查对象属性: getattr(ob, name)
        - 其次用 [] 检查: ob[name] 
        '''
        name = self.name
        op = self.op
        pattern = self.pattern

        if hasattr(ob, name):
            value = getattr(ob, name)
        else:
            try:
                value = ob[name]
            except KeyError:
                if '!' in op:
                    return True
                return False

        if isinstance(value, int):
            value = str(value)

        if op.startswith('#'):
            if isinstance(value, str):
                value = value.encode('latin1', 'replace')
            op = op[1:]
            value = binascii.hexlify(value)
            pattern = pattern.lower()

        if isinstance(value, bytes):
            value = value.decode('latin1', 'replace')

        if op == '=':
            return value == pattern
        elif op == '!=':
            return value != pattern
        elif op == '~=':
            return fnmatch.fnmatchcase(value, pattern)
        elif op == '!~=':
            return not fnmatch.fnmatchcase(value, pattern)
        else:
            assert False


class FilterGroup(list):
    def __init__(self, string=None):
        if string:
            self.parse(string)


class AndFilterGroup(FilterGroup):
    def parse(self, string):
        self.extend(map(Filter, string.split('&&')))

    def __repr__(self):
        return ' && '.join(map(str, self))

    def __call__(self, ob):
        for f in self:
            if not f(ob):
                return False

        return True


class OrFilterGroup(FilterGroup):
    def parse(self, string):
        self.extend(map(AndFilterGroup, string.split('||')))

    def __repr__(self):
        return ' || '.join(map(str, self))

    def __call__(self, ob):
        if not self:
            return True

        for f in self:
            if f(ob):
                return True

        return False


if __name__ == '__main__':
    f = OrFilterGroup('a=a && b=b || a=A && b=B')
    print(f)
    assert f(dict(a='a', b='b')) == True
    assert f(dict(a='A', b='B')) == True
    assert f(dict(a='a', b='B')) == False
    assert f(dict(a=1, b=2)) == False



# -*- coding: utf-8 -*-
import unittest
from pu import rcp


class Test(unittest.TestCase):
    def test_encode(self):
        self.assertEqual(
            rcp.encode(rcp.Notify('hello', ['tom', 'bob'], {})),
            ['N', 'hello', ['tom', 'bob'], {}])

        self.assertEqual(
            rcp.encode(rcp.Call(99, 'add', (1, 2), {})),
            ['C', 99, 'add', (1, 2), {}])

        self.assertEqual(
            rcp.encode(rcp.Return(99, 3)),
            ['R', 99, 3])

        self.assertEqual(
            rcp.encode(rcp.Error(99, 101, '协议错误', None)),
            ['E', 99, 101, '协议错误', None])

    def test_decode(self):
        self.assertEqual(
            rcp.decode(['N', 'hello', ['tom', 'bob'], {}]),
            rcp.Notify('hello', ['tom', 'bob'], {}))

        self.assertEqual(
            rcp.decode(['C', 99, 'add', (1, 2), {}]),
            rcp.Call(99, 'add', (1, 2), {}))

        self.assertEqual(
            rcp.decode(['R', 99, 3]),
            rcp.Return(99, 3))

        self.assertEqual(
            rcp.decode(['E', 99, 101, '协议错误', None]),
            rcp.Error(99, 101, '协议错误', None))


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main()



# -*- coding: utf-8 -*-
import unittest

from pu.manager import Manager


class Test(unittest.TestCase):
    def setUp(self):
        self.manager = Manager()

    def test_register_with_name(self):
        self.manager.register(1, 'ONE')
        self.assertEqual(self.manager.get('ONE'), 1)

    def test_register_function(self):
        @self.manager.register
        def f(): pass
        self.assertEqual(self.manager.get('f'), f)

    def test_register_class(self):
        @self.manager.register
        class C(): pass
        self.assertEqual(self.manager.get('C'), C)

    def test_named(self):
        self.manager.named('one', 'ONE')(1)
        self.assertEqual(self.manager.get('one'), 1)
        self.assertEqual(self.manager.get('ONE'), 1)

    def test_named_function(self):
        @self.manager.named('funcf', 'FUNCF')
        def f(): pass
        self.assertEqual(self.manager.get('funcf'), f)
        self.assertEqual(self.manager.get('FUNCF'), f)

    def test_named_class(self):
        @self.manager.named('clsc', 'CLSC')
        class C(): pass
        self.assertEqual(self.manager.get('clsc'), C)
        self.assertEqual(self.manager.get('CLSC'), C)

    def test_duplicate(self):
        self.manager.register(1, 'ONE')
        self.assertRaises(AssertionError,
            self.manager.register, '1', 'ONE')



# -*- coding: utf-8 -*-
from distutils.core import setup
import os.path
import sys

import pu


classifiers = [
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
    "Operating System :: OS Independent",
    "Topic :: Software Development :: Libraries :: Python Modules",
    ]


def read(fname):
    fname = os.path.join(os.path.dirname(__file__), fname)
    if sys.version > '3.0':
        content = open(fname, encoding='utf-8').read()
    else:
        content = open(fname).read().decode('utf-8')
    return content.strip()

def read_files(*fnames):
    return '\r\n\r\n\r\n'.join(map(read, fnames))

setup(
    name = 'pu',
    version=pu.version,
    packages = [
        'pu',
        'pu.aio',
        'pu.aio.protocols',
        'pu.pattern',
        'pu.misc',
        ],
    description = 'Python utils',
    long_description = read_files('README.rst', 'CHANGES.rst'),
    license = 'GNU Library or Lesser General Public License (LGPL)',
    author = 'yuxin',
    author_email = 'ycyuxin@gmail.com',
    url = 'https://github.com/huyx/pu',
    keywords = ['python', 'util'],
    classifiers = classifiers, 
    )



# -*- coding: utf-8 -*-
from ast import literal_eval
import re
import shlex


def guess(value):
    try:
        return literal_eval(value)
    except:
        pass

    try:
        if value.lower() in ('yes', 'true'):
            return True
        if value.lower() in ('no', 'false'):
            return False
    except:
        pass

    return value


patterns = (
    ('--', re.compile(r'^--$')),  # -- arg1 arg2
    ('--arg=value', re.compile(r'^--(.*)=(.*)$')),
    ('-a=value', re.compile(r'^-(.*)=(.*)$')),
    ('--noarg', re.compile(r'^--no(.*)$')),
    ('--arg.', re.compile(r'^--[^.]*\.$')),  # --arg.
    ('-a.', re.compile(r'^-.\.$')),  # -a.
    ('--arg', re.compile(r'^--[^.]*$')),  # --arg value | --arg
    ('-a', re.compile(r'^-.$')),  # -a value | -a
    ('-aNN', re.compile(r'^-(.)([0-9]+)$')),  # -a88
    ('-abc', re.compile(r'^-(.*)$')),  # -a -b -c
    )


def parse(args, *, lists=[], bools=[], strings=[], defaults={}, comments=True):
    '''分析命令行参数
    '''
    if isinstance(args, str):
        args = shlex.split(args, comments=comments)

    # 准备结果
    result = {}

    result['_'] = []

    # 列表选项处理
    for name in lists:
        result[name] = []

    # 定义设定选项的函数
    def set_arg(name, value, is_string=False):
        if not is_string and isinstance(value, str):
            value = guess(value)

        orig_value = result.get(name, None)

        if orig_value is not None:
            # 多个值自动转换成列表
            if isinstance(orig_value, list):
                orig_value.append(value)
            else:
                result[name] = [orig_value, value]
        else:
            result[name] = value

    while args:
        arg = args.pop(0)

        name = arg.lstrip('-')

        if name in bools:
            set_arg(name, True)
            continue

        if name in strings:
            assert args, '{} must follow a value'.format(arg)
            value = args.pop(0)
            set_arg(name, value, is_string=True)
            continue

        for cond, pattern in patterns:
            m = pattern.match(arg)
            if m:
                if cond == '--':
                    set_arg('--', args)
                    args = []
                    break
                elif cond in ('--arg=value', '-a=value'):
                    set_arg(*m.groups())
                elif cond == '--noarg':
                    set_arg(m.group(1), False)
                elif cond in ('--arg', '-a'):
                    if args:
                        value = args[0]
                        if value[0] != '-':
                            args.pop(0)
                            set_arg(name, value)
                        else:
                            set_arg(name, True)
                    else:
                        set_arg(name, True)
                elif cond in ('--arg.', '-a.'):
                    name = name[:-1]  # 去掉末尾的 .
                    set_arg(name, True)
                elif cond == '-aNN':
                    name, value = m.groups()
                    set_arg(name, int(value))
                elif cond == '-abc':
                    for n in name:
                        set_arg(n, True)
                break
        else:
            set_arg('_', arg)

    # 缺省值处理
    for name, value in defaults.items():
        if name not in result:
            result[name] = value

    return result


if __name__ == '__main__':
    import sys

    print(parse(sys.argv[1:]))



# -*- coding: utf-8 -*-
from .datatype import pretty_bytes
from .dictutil import DotOrderedDict
import socket
import sys


class Packet(DotOrderedDict):
    def __init__(self):
        super().__init__()

        self.type = self.__class__.__name__


class Ethernet(Packet):
    def __init__(self, ether_type, src, dst, data):
        super().__init__()

        self.ether_type = ether_type
        self.src = src
        self.dst = dst
        self.data = pretty_bytes(data)

    def __str__(self):
        return '{type} {ether_type:04x} {src} > {dst} {data:hex}'.format_map(self)


class IP(Packet):
    def __init__(self, sip, dip, proto, data):
        super().__init__()

        self.proto = proto
        self.sip = sip
        self.dip = dip
        self.data = pretty_bytes(data)

    def __str__(self):
        return '{type} {proto:04x} {sip} > {dip} {data:hex}'.format_map(self)


class ICMP(Packet):
    def __init__(self, sip, dip, data):
        super().__init__()

        self.sip = sip
        self.dip = dip
        self.data = pretty_bytes(data)

    def __str__(self):
        return '{type} {sip} > {dip} {data:hex}'.format_map(self)


class TCP(Packet):
    def __init__(self, sip, sport, dip, dport, flags, seq_no, data):
        super().__init__()

        self.sip = sip
        self.sport = sport
        self.dip = dip
        self.dport = dport
        self.flags = flags
        self.seq_no = seq_no
        self.data = pretty_bytes(data)

    def __str__(self):
        return '{type} {sip}:{sport} > {dip}:{dport} [{flags}] {seq_no:08x} {data:hex}'.format_map(self)


class UDP(Packet):
    def __init__(self, sip, sport, dip, dport, data):
        super().__init__()

        self.sip = sip
        self.sport = sport
        self.dip = dip
        self.dport = dport
        self.data = pretty_bytes(data)

    def __str__(self):
        return '{type} {sip}:{sport} > {dip}:{dport} {data:hex}'.format_map(self)


def create_raw_socket(interface=None):
    '''创建 RAW SOCKET

    网络接口:

    - Linux 平台: lo, eth0, eth1, ...
    - Windows 平台: IP 地址

    :param interface: 网络接口
    '''
    if sys.platform == 'linux':
        s = socket.socket(socket.AF_PACKET , socket.SOCK_RAW , socket.ntohs(0x0003))
        if interface:
            s.bind((interface, 0))
    elif sys.platform == 'win32':
        # 创建 raw socke
        s = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_IP)

        if not interface:
            # 获取 public network interface
            interface = socket.gethostbyname(socket.gethostname())
        s.bind((interface, 0))

        # 包括 IP 头
        s.setsockopt(socket.IPPROTO_IP, socket.IP_HDRINCL, 1)

        # 接收所有数据包
        s.ioctl(socket.SIO_RCVALL, socket.RCVALL_ON)
    else:
        assert False

    return s


def parse(data):
    '''分析收到的数据

    返回 None, IPPacket, 
    '''
    data_viewer = memoryview(data)

    if sys.platform == 'linux':
        src = ':'.join('%02x' % b for b in data_viewer[:6])
        dst = ':'.join('%02x' % b for b in data_viewer[6:12])
        ether_type = int.from_bytes(data[12:14], 'big')
        data_viewer = data_viewer[14:]

        # 只分析 IP 协议
        if ether_type != 0x0800:
            return Ethernet(ether_type, src, dst, data_viewer)
    else:
        data_viewer = data_viewer

    ip_header = data_viewer[:20]
    ip_data = data_viewer[20:]

    proto = ip_header[9]
    src_ip = socket.inet_ntoa(ip_header[12:16].tobytes())
    dst_ip = socket.inet_ntoa(ip_header[16:20].tobytes())

    if proto == 1:     # ICMP
        return ICMP(src_ip, dst_ip, ip_data)
    elif proto == 6:   # TCP
        tcp_header = ip_data
        data_offset = tcp_header[12] >> 4
        tcp_data = ip_data[data_offset*4:]

        src_port = int.from_bytes(tcp_header[0:2], 'big')
        dst_port = int.from_bytes(tcp_header[2:4], 'big')
        seq_no = int.from_bytes(tcp_header[4:8], 'big')

        flags = tcp_header[13]

        flags = ''.join((
            flags & 0x20 and 'U' or '',
            flags & 0x10 and 'A' or '',
            flags & 0x08 and 'P' or '',
            flags & 0x04 and 'R' or '',
            flags & 0x02 and 'S' or '',
            flags & 0x01 and 'F' or '',
            ))

        return TCP(src_ip, src_port, dst_ip, dst_port, flags, seq_no, tcp_data)

    elif proto == 17:  # UDP
        udp_header = ip_data
        udp_data = ip_data[8:]

        src_port = int.from_bytes(udp_header[0:2], 'big')
        dst_port = int.from_bytes(udp_header[2:4], 'big')

        return UDP(src_ip, src_port, dst_ip, dst_port, udp_data)
    else:
        return IP(src_ip, dst_ip, proto, ip_data)


def pcap(interface=None):
    s = create_raw_socket(interface)
    while True:
        data, _addr = s.recvfrom(65535)
        packet = parse(data)
        if packet:
            yield packet


if __name__ == '__main__':
    import argparse

    ARGS = argparse.ArgumentParser()

    ARGS.add_argument(
        '--interface', '-i', action='store', dest='interface',
        default=None, help='网络接口(Linux 下 lo, eth0 等, Windows 下是本机IP)')

    ARGS.add_argument(
        '--count', '-n', action='store', dest='count',
        type=int, default=5, help='收到指定数目的数据包后退出.')

    args = ARGS.parse_args()

    count = args.count
    for packet in pcap(args.interface):
        print(packet)
        if count:
            count -= 1
            if not count:
                break



# -*- coding: utf-8 -*-
import asyncio
import logging


logger = logging.getLogger(__name__)


class Client:
    def __init__(self, protocol_factory, loop=None):
        self.protocol_factory = protocol_factory
        self.loop = loop or asyncio.get_event_loop()

    @asyncio.coroutine
    def connect(self, host, port):
        self.host = host
        self.port = port

        yield from self._connect()

    @asyncio.coroutine
    def _connect(self):
        logger.info('连接 %s:%d', self.host, self.port)

        result = yield from self.loop.create_connection(
            self.protocol_factory, self.host, self.port)
        return result


class ReconnectingClient(Client):
    def __init__(self, protocol_factory, retry_delay=5, loop=None):
        super().__init__(protocol_factory, loop)

        self.retry_delay = retry_delay

    @asyncio.coroutine
    def _connect(self, delay=0):
        yield from asyncio.sleep(delay)

        while True:
            try:
                # create_connection 返回 transport, protocol
                _, protocol = yield from super()._connect()
                self._replace_connection_lost(protocol)
                break
            except Exception as e:
                logger.warning('连接 %s:%d 失败: %s', self.host, self.port, e)
                yield from asyncio.sleep(self.retry_delay)

    def _replace_connection_lost(self, protocol):
        # 截获 protocol 的 connection_lost 进行重连
        old_connection_lost = protocol.connection_lost

        def connection_lost(exc):
            logger.info('连接断开，重新连接: %s', exc)
            old_connection_lost(exc)
            asyncio.async(self._connect(self.retry_delay))

        protocol.connection_lost = connection_lost


if __name__ == '__main__':
    import sys

    logging.basicConfig(level='DEBUG')

    loop = asyncio.get_event_loop()

    if '-r' in sys.argv:
        client = ReconnectingClient(asyncio.Protocol)
    else:
        client = Client(asyncio.Protocol)

    loop.run_until_complete(client.connect('127.0.0.1', 9999))

    logger.info('连接完成，可以做自己的事了.')

    loop.run_forever()



# -*- coding: utf-8 -*-
'''lazydict
lazy 在这里有两个意思：

1. setdefault 时可以选择延迟生成对象
2. 提供一些方便地方法访问对象

d = lazydict()
d.setdefault('list', factory=list)
'''
from collections import Mapping, Sequence, MutableSequence, MutableMapping


def _smart_get(ob, key, default=None):
    if isinstance(ob, Sequence):
        return ob[int(key)]
    elif isinstance(ob, Mapping):
        return ob.get(key, default)
    else:
        return getattr(ob, key, default)


def smart_get(ob, key, default=None):
    if '.' not in key:
        return _smart_get(ob, key, default)

    first, _, other = key.partition('.')
    child = _smart_get(ob, first)

    assert child is not None, '{} 没有 {}.'.format(ob, first)

    return smart_get(child, other, default)


def _smart_set(ob, key, value):
    if isinstance(ob, MutableSequence):
        ob[int(key)] = value
    elif isinstance(ob, MutableMapping):
        ob[key] = value
    else:
        setattr(ob, key, value)


def smart_set(ob, key, value):
    if '.' not in key:
        return _smart_set(ob, key, value)

    first, _, other = key.partition('.')
    child = _smart_get(ob, first)

    assert child is not None, '{} 没有 {}.'.format(ob, first)

    return smart_set(child, other, value)


class lazydict(dict):
    def setdefault(self, key, default=None, *, factory=None):
        if key in self:
            return self[key]
        if factory:
            default = factory()
        return dict.setdefault(self, key, default)

    # 绑定 smart_get, smart_set
    smart_get = smart_get
    smart_set = smart_set
